import numpy as np
import copy as cp
from Orbaplaw import Integrals as eint
from . import OrbitalAlignment


def SpinAlignment(mo_mwfn, diagmat = False, diagmis = True):
	if mo_mwfn.Overlap.shape != tuple([mo_mwfn.getNumBasis()] * 2):
		mo_mwfn.Overlap = eint.PyscfOverlap(mo_mwfn, mo_mwfn)
	sno_mwfn = mo_mwfn.Clone()
	if sno_mwfn.Wfntype != 1:
		raise RuntimeError("Spin alignment can only be done on spin-unrestricted wavefunctions!")
	S = sno_mwfn.Overlap
	noccA = round(sno_mwfn.getNumElec(1))
	A = sno_mwfn.getCoefficientMatrix(1)
	eA = sno_mwfn.getEnergy(1)
	noccB = round(sno_mwfn.getNumElec(2))
	B = sno_mwfn.getCoefficientMatrix(2)
	C = np.zeros_like(sno_mwfn.getCoefficientMatrix(1))
	if noccA < noccB:
		raise RuntimeError("The first spin type must have no fewer electrons than the second!")
	A[:, :noccA], eA[:noccA] = OrbitalAlignment(A[:, :noccA], B[:, :noccB], S, eA[:noccA], diagmat, diagmis)
	A[noccA:, :] = 0
	eA[noccA:] = [0] * (len(eA) - noccA)
	sno_mwfn.setCoefficientMatrix(A, 1)
	sno_mwfn.setEnergy(eA, 1)
	return sno_mwfn



import numpy as np


def generatePrincipalInteractingOrbital(basis_indices_by_frag,P):

	# Obtaining NAOs in fragments
	bfA=basis_indices_by_frag[0]
	bfB=basis_indices_by_frag[1]

	# Principal interacting orbital
	Pab=P[np.ix_(bfA,bfB)]
	U,Sigma,VT=np.linalg.svd(Pab)
	T=np.zeros_like(P)
	T[np.ix_(bfA,bfA)]=U # NAO -> PIO
	T[np.ix_(bfB,bfB)]=VT.T
	I=np.zeros(P.shape[0])
	I[:len(Sigma)]=Sigma**2 # PIO-based bond index

	# Ordering PIOs in pairs
	TT=T.copy()
	II=I.copy()
	pair_info=[]
	pio=0
	nopair=set([i for i in range(P.shape[0])])
	for ipair in range(len(bfA)+len(bfB)):
		if ipair<len(bfA) and ipair<len(bfB):
			T[:,pio]=TT[:,bfA[ipair]]
			I[pio]=II[ipair]
			T[:,pio+1]=TT[:,bfB[ipair]]
			I[pio+1]=II[ipair]
			nopair-={bfA[ipair],bfB[ipair]}
			pair_info.append({"pimos":[pio,pio+1],"pios":[pio,pio+1]})
			pio+=2
		elif len(nopair)>0:
			ipio=nopair.pop()
			T[:,pio]=TT[:,ipio]
			I[pio]=II[ipio]
			pair_info.append({"pimos":[pio],"pios":[pio]})
			pio+=1
	Ppio=T.T@P@T # PIO-based density matrix
	N=np.diag(Ppio) # PIO population

	# Principal interacting molecular orbital
	Y=np.zeros_like(T) # PIO -> PIMO
	O=np.zeros_like(N) # Occupation of PIMO
	if len(basis_indices_by_frag)==2:
		for ipair in range(len(pair_info)):
			pips=pair_info[ipair]["pios"]
			Pblock=Ppio[np.ix_(pips,pips)]
			Oblock,Yblock=np.linalg.eigh(Pblock)
			O[pips]=Oblock[::-1]
			Y[np.ix_(pips,pips)]=np.fliplr(Yblock)
	
	return I,N,T,O,Y,pair_info

def PrincipalInteractingOrbital(nao_mwfn, nao_info, frags):
	basis_indices_by_center = nao_mwfn.Atom2BasisList()
	basis_indices_by_frag= []
	for frag in frags:
		basis_indices_this_fragment = []
		for icenter in frag:
			if icenter >= nao_mwfn.getNumCenters():
				raise RuntimeError("Atom index out of range!")
			basis_indices_this_fragment.extend(basis_indices_by_center[icenter])
		basis_indices_by_frag.append(basis_indices_this_fragment)
	pio_mwfn = nao_mwfn.Clone()
	pimo_mwfn = nao_mwfn.Clone()
	nbasis = nao_mwfn.getNumBasis()
	print("Principal interacting orbitals:")
	for spin in nao_mwfn.getSpins():
		C = nao_mwfn.getCoefficientMatrix(spin)
		P = None
		match spin:
			case 0:
				P = nao_info["NAO_density_matrix"]
			case 1:
				P = nao_info["NAO_alpha_density_matrix"]
			case 2:
				P = nao_info["NAO_beta_density_matrix"]
		I, N, T, O, Y, pair_info = generatePrincipalInteractingOrbital(basis_indices_by_frag, P)
		I *= 1 if spin == 0 else 2
		output = "Spin " + str(spin) + "\n"
		if len(basis_indices_by_frag) == 2: # Printing PIMO components
			for ipair in range(len(pair_info)):
				pimos = pair_info[ipair]["pimos"]
				pios = pair_info[ipair]["pios"]
				for pimo in pimos:
					output += f"PIMO_{pimo + (nbasis if spin == 2 else 0)} ({O[pimo]: .3f}, {I[pimo]:.3f}) ="
					for pio in pios:
						output += f"  {Y[pio,pimo]: .3f} * PIO_{pio+ (nbasis if spin==2 else 0)} ({N[pio]:.3f})"
					output+="\n"
		print(output)
		pio_mwfn.setOccupation(N, spin)
		pio_mwfn.setEnergy(I, spin)
		pio_mwfn.setCoefficientMatrix(C @ T, spin)
		pimo_mwfn.setOccupation(O, spin)
		pimo_mwfn.setEnergy(I, spin)
		pimo_mwfn.setCoefficientMatrix(C @ T @ Y, spin)
	print("Warning: The indeces of orbitals and fragments above start from 0!")
	return pio_mwfn, pimo_mwfn



import numpy as np
from Orbaplaw import Integrals as eint
from . import OrbitalAlignment


def FragmentAlignment(big_mwfn, small_mwfn_list, diagmat = False, diagmis = True):
	famo_mwfn = big_mwfn.Clone()
	overlap_matrices = []
	for small_mwfn in small_mwfn_list:
		overlap_matrices.append(eint.PyscfOverlap(big_mwfn, small_mwfn))
	S = np.hstack(overlap_matrices)
	for spin in big_mwfn.getSpins():
		noccA = round(big_mwfn.getNumElec(spin) / (2 if spin == 0 else 1))
		A = big_mwfn.getCoefficientMatrix(spin)[:, :noccA]
		epsilon = big_mwfn.getEnergy(spin)[:noccA]
		nbasisB = 0
		B_basis_heads = []
		B_basis_tails = []
		noccB = 0
		B_orbital_heads = []
		B_orbital_tails = []
		for small_mwfn in small_mwfn_list:
			B_basis_heads.append(nbasisB)
			nbasisB += small_mwfn.getNumBasis()
			B_basis_tails.append(nbasisB)
			B_orbital_heads.append(noccB)
			noccB += round(small_mwfn.getNumElec(spin) / (2 if spin == 0 else 1))
			B_orbital_tails.append(noccB)
		B = np.zeros([nbasisB, noccB])
		for i, small_mwfn in enumerate(small_mwfn_list):
			B_basis_head = B_basis_heads[i]
			B_basis_tail = B_basis_tails[i]
			B_orbital_head = B_orbital_heads[i]
			B_orbital_tail = B_orbital_tails[i]
			noccB = B_orbital_tail - B_orbital_head
			B[B_basis_head : B_basis_tail, B_orbital_head : B_orbital_tail] = small_mwfn.getCoefficientMatrix(spin)[:, :noccB]
		C = big_mwfn.getCoefficientMatrix(spin)
		E = big_mwfn.getEnergy(spin)
		C[:, :noccA], E[:noccA] = OrbitalAlignment(A, B, S, epsilon, diagmat, diagmis)
		famo_mwfn.setCoefficientMatrix(C, spin)
		famo_mwfn.setEnergy(E, spin)
	return famo_mwfn



import numpy as np
import scipy.linalg as sl
import scipy.optimize as so
import copy as cp

def LowdinCharge(C,S,basis_indices_by_center):
    nbasis=C.shape[0]
    norbitals=C.shape[1]
    assert (nbasis,nbasis)==S.shape,"Dimensions of coefficient and overlap matrices do not match!"
    S12C=sl.sqrtm(S)@C
    Qs=[]
    for basis_indices in basis_indices_by_center:
        Qa=np.zeros([norbitals,norbitals])
        for i in range(norbitals):
            for j in range(norbitals):
                for mu in basis_indices:
                    Qa[i,j]+=S12C[mu,i]*S12C[mu,j]
        Qs.append(Qa)
    return Qs

def Lower2Skew(x):
    nrows=int((1+np.sqrt(1+8*len(x)))/2)
    X=np.zeros([nrows,nrows])
    i=0
    for k in range(nrows):
        for l in range(k):
            X[k,l]=x[i]
            X[l,k]=-x[i]
            i+=1
    return X

def Skew2Lower(X):
    nrows=X.shape[0]
    x=[]
    for k in range(nrows):
        for l in range(k):
            x.append(X[k,l])
    return x

def PM_function(x,*args):
    X=Lower2Skew(x)
    C,S,basis_indices_by_center,charge_type,reverse=args
    Qs=charge_type(C@sl.expm(-X),S,basis_indices_by_center)
    function=0
    for Q in Qs:
        for i in range(Q.shape[0]):
            function+=Q[i,i]**2
    return function if reverse else -function

def PM_gradient(x,*args):
    X=Lower2Skew(x)
    C,S,basis_indices_by_center,charge_type,reverse=args
    Qs=charge_type(C@sl.expm(-X),S,basis_indices_by_center)
    gradient=np.array([])
    for k in range(Qs[0].shape[0]):
        for l in range(k):
            element=0
            for Q in Qs:
                element+=4*(Q[k,k]-Q[l,l])*Q[k,l]
            gradient=np.append(gradient,element)
    return gradient if reverse else -gradient


def generatePipekMezey(C,S,basis_indices_by_center,charge_type,reverse):
    norbitals=C.shape[1]
    U=np.eye(norbitals)
    max_gamma_error=114514
    while max_gamma_error>0.5: # Jacobi sweep
        max_gamma_error=0
        for k in range(norbitals):
            for l in range(k):
                Ast=0
                Bst=0
                S12C=sl.sqrtm(S)@C@U
                for basis_indices in basis_indices_by_center:
                    Qkk=0
                    Qll=0
                    Qkl=0
                    for mu in basis_indices:
                        Qll+=S12C[mu,l]*S12C[mu,l]
                        Qkk+=S12C[mu,k]*S12C[mu,k]
                        Qkl+=S12C[mu,k]*S12C[mu,l]
                    Ast+=Qkl**2-0.25*(Qkk-Qll)**2
                    Bst+=Qkl*(Qkk-Qll)
                if np.sqrt(Ast**2+Bst**2)<1e-10:
                    continue
                sin4a=Bst/np.sqrt(Ast**2+Bst**2)
                cos4a=-Ast/np.sqrt(Ast**2+Bst**2)
                four_a=(np.arccos(cos4a) if sin4a>0 else 2*np.pi-np.arccos(cos4a)) if reverse else np.sign(sin4a)*np.arccos(cos4a)
                a=four_a/4.
                gamma=a-0.25*np.pi if reverse else a
                gamma_error=(0.25*np.pi-a)**2 if reverse else abs((gamma-0.5*np.pi)*gamma)
                max_gamma_error=max_gamma_error if max_gamma_error > gamma_error else gamma_error
                U[:,[l,k]]=U[:,[l,k]]@np.array([[np.cos(gamma),np.sin(gamma)],[-np.sin(gamma),np.cos(gamma)]])
    x0=[0 for i in range(int(norbitals*(norbitals-1)/2))]
    PM=so.minimize(PM_function,x0,args=(C@U,S,basis_indices_by_center,charge_type,reverse),method="trust-ncg",jac=PM_gradient,hess="2-point")
    X=Lower2Skew(PM.x)
    return U@sl.expm(-X)

def PipekMezey(mo_mwfn,occ=True,vir=False,charge_type=LowdinCharge):
    S=mo_mwfn.Overlap_matrix
    basis_indices_by_center=mo_mwfn.getBasisIndexByCenter()
    pm_mwfn=cp.deepcopy(mo_mwfn)
    print("Pipek-Mezey localization:")
    if mo_mwfn.Wfntype==0 or mo_mwfn.Wfntype==1:
        for spin in ([0] if mo_mwfn.Wfntype==0 else [1,2]):
            print("Spin "+str(spin))
            C=mo_mwfn.getCoefficientMatrix(spin)
            nocc=mo_mwfn.Naelec if spin==1 else mo_mwfn.Nbelec
            if occ:
                print("Localizing occupied orbitals")
                Uocc=generatePipekMezey(C[:,:nocc],S,basis_indices_by_center,charge_type,0)
                C[:,:nocc]=C[:,:nocc]@Uocc
            if vir:
                print("Localizing virtual orbitals")
                Uvir=generatePipekMezey(C[:,nocc:],S,basis_indices_by_center,charge_type,0)
                C[:,nocc:]=C[:,nocc:]@Uvir
            pm_mwfn.setCoefficientMatrix(spin,C)
            pm_mwfn.setEnergy(spin,[0 for i in range(mo_mwfn.getNumIndBasis())])
    return pm_mwfn



import numpy as np
import copy as cp
from Orbaplaw import Integrals as eint
from Orbaplaw import Population as pop
from Orbaplaw import Miscellany as mis
from . import PipekMezey
from . import FosterBoys
from . import Fock
from . import Orbitalet


def Localizer(mo_mwfn, method = "PipekMezey-Lowdin", space = "occ"):

	loc_mwfn = mo_mwfn.Clone()
	norbitals = loc_mwfn.getNumIndBasis()
	for spin in loc_mwfn.getSpins():
		nocc = round(loc_mwfn.getNumElec(spin))
		if spin == 0:
			nocc //= 2
		orbital_range = []
		if type(space) is str:
			if space.upper() == "OCC":
				orbital_range = list(range(nocc))
			if space.upper() == "VIR":
				orbital_range = list(range(nocc, norbitals))
			if space.upper() == "MIX":
				orbital_range = list(range(norbitals))
		elif type(space) is list:
			orbital_range = space
		format_range = mis.FormatRange(orbital_range)
		Cold = loc_mwfn.getCoefficientMatrix(spin)[:, orbital_range]
		Cnew = Cold.copy()

		if "PM" in method.upper() or "PIPEK" in method.upper() or "MEZEY" in method.upper():
			S = loc_mwfn.Overlap
			if loc_mwfn.Overlap.shape != tuple([loc_mwfn.getNumBasis()] * 2):
				S = eint.PyscfOverlap(loc_mwfn, loc_mwfn)
			basis_indices_by_center = loc_mwfn.Atom2BasisList()
			charge_type = ""
			Qrefs = []
			if "LOWDIN" in method.upper():
				charge_type = "Lowdin"
				Qrefs = pop.Lowdin(Cold, S, basis_indices_by_center)
			elif "MULLIKEN" in method.upper():
				charge_type = "Mulliken"
				Qrefs = pop.Mulliken(Cold, S, basis_indices_by_center)
			else:
				raise RuntimeError("Unrecognized charge type!")
			print("Pipek-Mezey localization (%s) on Spin %d Orbitals %s:" % ( charge_type, spin, format_range) )
			occ_all = mo_mwfn.getOccupation(spin)
			occ = [occ_all[i] for i in orbital_range]
			mix = len(set(occ)) > 1 # Different occupation values among the orbitals indicate that occupied and unoccupied orbitals are mixed.
			if mix:
				raise RuntimeError("Fractionally occupied orbitals and mixing occupied and virtual orbitals in Pipek-Mezey localization is not supported!")
			Cnew = Cold @ PipekMezey(Qrefs)

		elif "FB" in method.upper() or "FOSTER" in method.upper() or "BOYS" in method.upper():
			print("Foster-Boys localization on Spin %d Orbitals %s:" % (spin, format_range))
			X, Y, Z = eint.PyscfDipole(loc_mwfn, loc_mwfn)
			XX, _, _, YY, _, ZZ = eint.PyscfQuadrupole(loc_mwfn, loc_mwfn)
			Waos = [ X, Y, Z ]
			Wrefs = [ Cold.T @ Wao @ Cold for Wao in Waos ]
			Wao2Sum = - XX - YY - ZZ
			W2refSum = Cold.T @ Wao2Sum @ Cold
			Cnew = Cold @ FosterBoys(Wrefs, W2refSum)

		elif "FOCK" in method.upper(): # Must start from CMOs
			print("Fock localization on Spin %d Orbitals %s:" % (spin, format_range))
			Eref = np.array(mo_mwfn.getEnergy(spin))[orbital_range]
			Cnew = Cold @ Fock(Eref)

		elif "OL" in method.upper() or "ORBITALET" in method.upper(): # Must start from CMOs
			gamma_e = 0.7959
			for word in method.split('#'):
				try:
					gamma_e = float(word)
					break
				except ValueError:
					continue
			if gamma_e <= 0 or gamma_e >= 1:
				raise RuntimeError("Gamma_e must be in (0, 1)!")
			print("Orbitalet localization (Gamma_e = %f) on Spin %d Orbitals %s:" % (gamma_e, spin, format_range))
			X, Y, Z = eint.PyscfDipole(loc_mwfn, loc_mwfn)
			XX, _, _, YY, _, ZZ = eint.PyscfQuadrupole(loc_mwfn, loc_mwfn)
			Waos = [ X, Y, Z ]
			Wrefs = [ Cold.T @ Wao @ Cold for Wao in Waos ]
			Wao2Sum = - XX - YY - ZZ
			W2refSum = Cold.T @ Wao2Sum @ Cold
			Eref = loc_mwfn.getEnergy(spin)[orbital_range]
			Cnew = Cold @ Orbitalet(Wrefs, W2refSum, Eref, gamma_e)

		C = loc_mwfn.getCoefficientMatrix(spin)
		C[:, orbital_range] = Cnew
		loc_mwfn.setCoefficientMatrix(C, spin)

	return loc_mwfn



import numpy as np
import scipy.linalg as sl
from Orbaplaw import Optimization as opt
from Orbaplaw import Population as pop


def PM_func(U,C0,S,basis_indices_by_center,charge_type):
    C=C0
    if U is not None:
        C=C0@U
    Qs=None
    if charge_type=="Lowdin":
        Qs=pop.Lowdin_func(C,S,basis_indices_by_center)
    L=0
    for Qa in Qs:
        L+=np.sum(np.diag(Qa)**2)
    return L

def PM_jac(U,C0,S,basis_indices_by_center,charge_type):
    C=C0@U
    Qs=None
    QrrUkrs=None
    if charge_type=="Lowdin":
        Qs=pop.Lowdin_func(C,S,basis_indices_by_center)
        QrrUkrs=pop.Lowdin_jac(C,C0,S,basis_indices_by_center)
    Gamma=np.zeros_like(U)
    for Qa,QrrUkra in zip(Qs,QrrUkrs):
        Gamma+=QrrUkra*np.diag(Qa)
    Gamma*=2 # Euclidean derivative
    return Gamma@U.T-U@Gamma.T # Riemannian derivative

def PM_conv(x,f,g,xlast,flast,glast):
    return np.max(np.abs(g))<1e-3 or abs(f-flast)<5e-6

def oldPipekMezey(C0,S,basis_indices_by_center,charge_type,conv):
    func=lambda x:PM_func(x,C0,S,basis_indices_by_center,charge_type)
    jac=lambda x:PM_jac(x,C0,S,basis_indices_by_center,charge_type)
    conv_=PM_conv if conv is None else conv
    return opt.Lehtola(C0,0.001,func,jac,4,conv_)

import Maniverse as mv

def PipekMezey(Qrefs):
	M = mv.Orthogonal(np.eye(Qrefs[0].shape[1]), True)
	def func(U, order):
		Qdiags = [ np.diag(np.diag(U.T @ Qref @ U)) for Qref in Qrefs ]
		L = 0
		for Qdiag in Qdiags:
			L -= np.linalg.norm(np.diag(Qdiag)) ** 2
		Ge = np.zeros_like(U)
		for Qref, Qdiag in zip(Qrefs, Qdiags):
			Ge += Qref @ U @ Qdiag
		Ge *= -4
		QrefUs = [ Qref @ U for Qref in Qrefs ]
		def He(v):
			return np.zeros_like(v)
		if order == 2:
			def He(v):
				hess1 = np.zeros_like(v)
				hess2 = np.zeros_like(v)
				for Qref, Qdiag, QrefU in zip(Qrefs, Qdiags, QrefUs):
					hess1 += Qref @ v @ Qdiag
					hess2 += QrefU @ np.diag(np.diag(QrefU.T @ v))
				return - 4 * hess1 - 8 * hess2
		return L, Ge, He
	L = 0
	tr_setting = mv.TrustRegionSetting()
	tol0 = 1e-8 * M.getDimension()
	tol1 = 1e-6 * M.getDimension()
	tol2 = 10
	mv.TrustRegion(
			func, tr_setting, (tol0, tol1, tol2),
			0.001, 1, 1000, L, M, 1
	)
	return M.P



from Orbaplaw.WaveFunction import MultiWaveFunction
MultiWaveFunction=MultiWaveFunction.MultiWaveFunction

from Orbaplaw.WaveFunction import QeWaveFunction
QeWaveFunction=QeWaveFunction.QeWaveFunction

from Orbaplaw.WaveFunction import Cube
Cube=Cube.Cube



from Orbaplaw.Population import Lowdin
Lowdin_func = Lowdin.Lowdin_func
Lowdin_jac = Lowdin.Lowdin_jac
Lowdin = Lowdin.Lowdin

from Orbaplaw.Population import PopulationAnalyzer
PopulationAnalyzer=PopulationAnalyzer.PopulationAnalyzer



from Orbaplaw import WaveFunction
from Orbaplaw import Population
from Orbaplaw import Localization
from Orbaplaw import NaturalBondOrbitalMethods
from Orbaplaw import Integrals
from Orbaplaw import Miscellany



import numpy as np
import scipy.linalg as sl
from Orbaplaw import Integrals as eint

def generateNaturalAtomicOrbital(shell_indices_by_center,basis_indices_by_shell,basis_indices_by_center,angulars,D,S,minimal_shells):
	P=S@D@S

	# 2: Intraatomic orthogonalization

	# 2a: Transformation from Cartesian to pure d, f, g AOs.
	# I do not want to code this. Tell users to use pure AOs only.

	# 2b: Partitioning and symmetry averaging of P and S.
	W=np.zeros(P.shape[0])
	N=np.zeros_like(P)
	for shell_indices_this_center in shell_indices_by_center:
		angulars_this_center=[angulars[shell_index] for shell_index in shell_indices_this_center]
		angulars_this_center_set=set(angulars_this_center)
		for angular in angulars_this_center_set:
			shell_indices=[shell_index for shell_index in shell_indices_this_center if angulars[shell_index]==angular]
			basis_index_heads=[basis_indices_by_shell[shell_index][0] for shell_index in shell_indices]
			PAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			SAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			for m in range(2*abs(angular)+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				PAl+=P[np.ix_(basis_indices,basis_indices)]
				SAl+=S[np.ix_(basis_indices,basis_indices)]
			PAl/=2*abs(angular)+1
			SAl/=2*abs(angular)+1

	# 2c: Formation of pre-NAOs.
			OL=np.linalg.inv(sl.sqrtm(SAl))
			PAlL=OL.T@PAl@OL
			w,NL=sl.eigh(PAlL)
			for m in range(2*abs(angular)+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				W[np.ix_(basis_indices)]=w
				N[np.ix_(basis_indices,basis_indices)]=OL@NL

	# 3: Initial division between orthogonal valence and Rydberg AO spaces.
	
	# 3a: Selection of NMB orbitals.
	WW=W.copy()
	NN=N.copy()
	W=np.array([])
	N=[]
	natoms=len(minimal_shells)
	basis_indices_nmb=[]
	basis_indices_nrb=[]
	jbasis=0
	for iatom in range(natoms):
		W_nmb_atom=np.array([])
		N_nmb_atom=[]
		old_shell_indices_nmb=[]
		old_basis_indices_nmb=[]
		W_nrb_atom=np.array([])
		N_nrb_atom=[]
		old_shell_indices_nrb=[]
		old_basis_indices_nrb=[]
		for l in range(len(minimal_shells[iatom])):
			shell_indices=[]
			ws=[]
			for shell_index in shell_indices_by_center[iatom]:
				if abs(angulars[shell_index])==l:
					shell_indices.append(shell_index)
					ws.append(WW[basis_indices_by_shell[shell_index][0]])
			args=np.argsort(ws)[::-1]
			for p in range(len(args)):
				if p<minimal_shells[iatom][l]: # NMB
					for m in range(2*l+1):
						W_nmb_atom=np.append(W_nmb_atom,WW[basis_indices_by_shell[shell_indices[args[p]]][0]+m])
						N_nmb_atom.append(NN[:,basis_indices_by_shell[shell_indices[args[p]]][0]+m])
					old_shell_indices_nmb.append(shell_indices[args[p]])
				else: # NRB
					for m in range(2*l+1):
						W_nrb_atom=np.append(W_nrb_atom,WW[basis_indices_by_shell[shell_indices[args[p]]][0]+m])
						N_nrb_atom.append(NN[:,basis_indices_by_shell[shell_indices[args[p]]][0]+m])
					old_shell_indices_nrb.append(shell_indices[args[p]])
		basis_indices_nmb_this_atom=[]
		basis_indices_nrb_this_atom=[]
		for old_shell_index in old_shell_indices_nmb:
			basis_indices_nmb_this_atom_this_shell=[]
			for m in range(2*abs(angulars[old_shell_index])+1):
				basis_indices_nmb_this_atom_this_shell.append(jbasis)
				jbasis+=1
			basis_indices_nmb_this_atom.append(basis_indices_nmb_this_atom_this_shell)
		for old_shell_index in old_shell_indices_nrb:
			basis_indices_nrb_this_atom_this_shell=[]
			for m in range(2*abs(angulars[old_shell_index])+1):
				basis_indices_nrb_this_atom_this_shell.append(jbasis)
				jbasis+=1
			basis_indices_nrb_this_atom.append(basis_indices_nrb_this_atom_this_shell)
		basis_indices_nmb.append(basis_indices_nmb_this_atom)
		basis_indices_nrb.append(basis_indices_nrb_this_atom)
		W_atom=np.append(W_nmb_atom,W_nrb_atom)
		W=np.append(W,W_atom)
		N_nmb_atom=np.array(N_nmb_atom)
		N_nrb_atom=np.array(N_nrb_atom)
		N.extend(N_nmb_atom)
		N.extend(N_nrb_atom)
	N=np.array(N).T

	# 4a: Weighted interatomic orthogonalization within NMB.
	all_basis_indices_nmb=sum(sum(basis_indices_nmb,[]),[]) # Flattening a list of lists of lists.
	all_basis_indices_nrb=sum(sum(basis_indices_nrb,[]),[])
	Spnao=N.T@S@N
	Snmb=Spnao[np.ix_(all_basis_indices_nmb,all_basis_indices_nmb)]
	Wnmb=np.diag(W[np.ix_(all_basis_indices_nmb)])
	Ownmb=Wnmb@np.linalg.inv(sl.sqrtm(Wnmb@Snmb@Wnmb))
	N[:,np.ix_(all_basis_indices_nmb)]=N[:,np.ix_(all_basis_indices_nmb)]@Ownmb
	Spnao=N.T@S@N

	# 3b: Schmidt interatomic orthogonalization of NRB to NMB orbitals.
	for a in range(Spnao.shape[0]):
		for r in all_basis_indices_nrb:
			for m in all_basis_indices_nmb:
				N[a,r]-=N[a,m]*Spnao[m,r]
	Spnao=N.T@S@N

	# 3c: Restoration of natural character of the NRB.
	W.setflags(write=True)
	Ppnao=N.T@P@N
	Nryd=np.zeros_like(N)
	Nryd[np.ix_(all_basis_indices_nmb,all_basis_indices_nmb)]=np.eye(len(all_basis_indices_nmb))
	for basis_indices_this_atom in basis_indices_nrb:
		lmax=int((max(map(len,basis_indices_this_atom))-1)/2) if len(basis_indices_this_atom)>0 else -2 # Skipping this part if a minimal basis is used.
		for l in range(lmax+1):
			basis_index_heads=[]
			for basis_index_this_atom_this_shell in basis_indices_this_atom:
				if 2*l+1==len(basis_index_this_atom_this_shell):
					basis_index_heads.append(basis_index_this_atom_this_shell[0])
			PAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			SAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			for m in range(2*l+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				PAl+=Ppnao[np.ix_(basis_indices,basis_indices)]
				SAl+=Spnao[np.ix_(basis_indices,basis_indices)]
			PAl/=2*l+1
			SAl/=2*l+1
			OL=np.linalg.inv(sl.sqrtm(SAl))
			PAlL=OL.T@PAl@OL
			w,NL=sl.eigh(PAlL)
			for m in range(2*l+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				W[np.ix_(basis_indices)]=w[::-1]
				Nryd[np.ix_(basis_indices,basis_indices)]=(OL@NL)[:,::-1]
	N=N@Nryd
	Spnao=N.T@S@N


	# 4: Formation of the final NAO set
	# 4a: Weighted interatomic orthogonalization within NRB.
	W=np.diag(W)
	Ow=np.real(W@np.linalg.inv(sl.sqrtm(W@Spnao@W)))
	N=N@Ow
	Spnao=N.T@S@N
	W=np.diag(W)
	'''
	print(sl.norm(Spnao[np.ix_(all_basis_indices_nmb,all_basis_indices_nmb)]-np.eye(len(all_basis_indices_nmb))))
	print(np.diag(Spnao[np.ix_(all_basis_indices_nrb,all_basis_indices_nrb)]))
	print(sl.norm(Spnao[np.ix_(all_basis_indices_nrb,all_basis_indices_nrb)]-np.eye(len(all_basis_indices_nrb))))
	print(sl.norm(Spnao[np.ix_(all_basis_indices_nmb,all_basis_indices_nrb)]))
	'''

	# 4b: Restoration of natural character of the NAOs.
	W.setflags(write=True)
	Ppnao=N.T@P@N
	Nred=np.zeros_like(N)
	basis_indices_both=[]
	for basis_indices_nmb_this_atom,basis_indices_nrb_this_atom in zip(basis_indices_nmb,basis_indices_nrb):
		basis_indices_both.append(basis_indices_nmb_this_atom+basis_indices_nrb_this_atom)
	for basis_indices_this_atom in basis_indices_both:
		lmax=int((max(map(len,basis_indices_this_atom))-1)/2)
		for l in range(lmax+1):
			basis_index_heads=[]
			for basis_index_this_atom_this_shell in basis_indices_this_atom:
				if l==(len(basis_index_this_atom_this_shell)-1)/2:
					basis_index_heads.append(basis_index_this_atom_this_shell[0])
			PAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			SAl=np.zeros([len(basis_index_heads),len(basis_index_heads)])
			for m in range(2*l+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				PAl+=Ppnao[np.ix_(basis_indices,basis_indices)]
				SAl+=Spnao[np.ix_(basis_indices,basis_indices)]
			PAl/=2*l+1
			SAl/=2*l+1
			OL=np.linalg.inv(sl.sqrtm(SAl))
			PAlL=OL.T@PAl@OL
			w,NL=sl.eigh(PAlL)
			for m in range(2*l+1):
				basis_indices=[basis_index+m for basis_index in basis_index_heads]
				W[np.ix_(basis_indices)]=w[::-1]
				Nred[np.ix_(basis_indices,basis_indices)]=(OL@NL)[:,::-1]
	N=N@Nred

	# Computing population of NAOs
	P_nao=N.T@P@N
	W=np.diag(P_nao)

	return basis_indices_nmb,basis_indices_nrb,W,N

def MinimalShells(an,nc): # an - Atomic number; nc - Nuclear charge
	match an:
		case 1:
			return [1,0,0,0,0,0,0] #  H
		case 2:
			return [1,0,0,0,0,0,0] # He
		case 3:
			return [2,0,0,0,0,0,0] # Li
		case 4:
			return [2,0,0,0,0,0,0] # Be
		case 5:
			return [2,1,0,0,0,0,0] #  B
		case 6:
			return [2,1,0,0,0,0,0] #  C
		case 7:
			return [2,1,0,0,0,0,0] #  N
		case 8:
			return [2,1,0,0,0,0,0] #  O
		case 9:
			return [2,1,0,0,0,0,0] #  F
		case 10:
			return [2,1,0,0,0,0,0] # Ne
		case 11:
			return [3,1,0,0,0,0,0] # Na
		case 12:
			return [3,1,0,0,0,0,0] # Mg
		case 13:
			return [3,2,0,0,0,0,0] # Al
		case 14:
			return [3,2,0,0,0,0,0] # Si
		case 15:
			match nc:
				case 5:
					return [1,1,0,0,0,0]
			return [3,2,0,0,0,0,0] #  P
		case 16:
			return [3,2,0,0,0,0,0] #  S
		case 17:
			return [3,2,0,0,0,0,0] # Cl
		case 18:
			return [3,2,0,0,0,0,0] # Ar
		case 19:
			match nc:
				case 9:
					return [2,1,0,0,0,0,0]
			return [4,2,0,0,0,0,0] #  K
		case 20:
			return [4,2,0,0,0,0,0] # Ca
		case 21:
			match nc:
				case 11:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Sc
		case 22:
			match nc:
				case 12:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Ti
		case 23:
			match nc:
				case 13:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] #  V
		case 24:
			match nc:
				case 14:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Cr
		case 25:
			match nc:
				case 15:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Mn
		case 26:
			match nc:
				case 16:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Fe
		case 27:
			match nc:
				case 17:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Co
		case 28:
			match nc:
				case 18:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Ni
		case 29:
			match nc:
				case 19:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Cu
		case 30:
			match nc:
				case 20:
					return [2,1,1,0,0,0,0]
			return [4,2,1,0,0,0,0] # Zn
		case 31:
			return [4,3,1,0,0,0,0] # Ga
		case 32:
			return [4,3,1,0,0,0,0] # Ge
		case 33:
			match nc:
				case 5:
					return [1,1,0,0,0,0,0]
			return [4,3,1,0,0,0,0] # As
		case 34:
			return [4,3,1,0,0,0,0] # Se
		case 35:
			return [4,3,1,0,0,0,0] # Br
		case 36:
			return [4,3,1,0,0,0,0] # Kr
		case 37:
			match nc:
				case 9:
					return [2,1,0,0,0,0,0]
			return [5,3,1,0,0,0,0] # Rb
		case 38:
			match nc:
				case 10:
					return [2,1,0,0,0,0,0]
			return [5,3,1,0,0,0,0] # Sr
		case 39:
			return [5,3,2,0,0,0,0] #  Y
		case 40:
			return [5,3,2,0,0,0,0] # Zr
		case 41:
			return [5,3,2,0,0,0,0] # Nb
		case 42:
			return [5,3,2,0,0,0,0] # Mo
		case 43:
			return [5,3,2,0,0,0,0] # Tc
		case 44:
			match nc:
				case 16:
					return [2,1,1,0,0,0,0]
			return [5,3,2,0,0,0,0] # Ru
		case 45:
			return [5,3,2,0,0,0,0] # Rh
		case 46:
			match nc:
				case 18:
					return [2,1,1,0,0,0,0]
			return [5,3,2,0,0,0,0] # Pd
		case 47:
			return [5,3,2,0,0,0,0] # Ag
		case 48:
			return [5,3,2,0,0,0,0] # Cd
		case 49:
			return [5,4,2,0,0,0,0] # In
		case 50:
			match nc:
				case 4:
					return [1,1,0,0,0,0,0]
				case 22:
					return [2,2,1,0,0,0,0]
			return [5,4,2,0,0,0,0] # Sn
		case 51:
			match nc:
				case 5:
					return [1,1,0,0,0,0,0]
				case 23:
					return [2,2,1,0,0,0,0]
			return [5,4,2,0,0,0,0] # Sb
		case 52:
			return [5,4,2,0,0,0,0] # Te
		case 53:
			match nc:
				case 25:
					return [2,2,1,0,0,0,0]
			return [5,4,2,0,0,0,0] #  I
		case 54:
			return [5,4,2,0,0,0,0] # Xe
		case 55:
			return [6,4,2,0,0,0,0] # Cs
		case 56:
			return [6,4,2,0,0,0,0] # Ba
		case 57:
			return [6,4,3,0,0,0,0] # La
		case 58:
			return [6,4,3,1,0,0,0] # Ce
		case 59:
			return [6,4,2,1,0,0,0] # Pr
		case 60:
			return [6,4,2,1,0,0,0] # Nd
		case 61:
			return [6,4,2,1,0,0,0] # Pm
		case 62:
			return [6,4,2,1,0,0,0] # Sm
		case 63:
			return [6,4,2,1,0,0,0] # Eu
		case 64:
			return [6,4,3,1,0,0,0] # Gd
		case 65:
			return [6,4,2,1,0,0,0] # Tb
		case 66:
			return [6,4,2,1,0,0,0] # Dy
		case 67:
			return [6,4,2,1,0,0,0] # Ho
		case 68:
			return [6,4,2,1,0,0,0] # Er
		case 69:
			return [6,4,2,1,0,0,0] # Tm
		case 70:
			return [6,4,2,1,0,0,0] # Yb
		case 71:
			match nc:
				case 11:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Lu
		case 72:
			match nc:
				case 12:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Hf
		case 73:
			match nc:
				case 13:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Ta
		case 74:
			match nc:
				case 14:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] #  W
		case 75:
			match nc:
				case 15:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Re
		case 76:
			match nc:
				case 16:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Os
		case 77:
			match nc:
				case 17:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Ir
		case 78:
			match nc:
				case 18:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Pt
		case 79:
			match nc:
				case 19:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Au
		case 80:
			match nc:
				case 20:
					return [2,1,1,0,0,0,0]
			return [6,4,3,1,0,0,0] # Hg
		case 81:
			return [6,5,3,1,0,0,0] # Tl
		case 82:
			return [6,5,3,1,0,0,0] # Pb
		case 83:
			match nc:
				case 5:
					return [1,1,0,0,0,0,0]
				case 23:
					return [2,2,1,0,0,0,0]
			return [6,5,3,1,0,0,0] # Bi
		case 84:
			return [6,5,3,1,0,0,0] # Po
		case 85:
			return [6,5,3,1,0,0,0] # At
		case 86:
			return [6,5,3,1,0,0,0] # Rn
		case 87:
			return [7,5,3,1,0,0,0] # Fr
		case 88:
			return [7,5,3,1,0,0,0] # Ra
		case 89:
			return [7,5,4,1,0,0,0] # Ac
		case 90:
			match nc:
				case 12:
					return [2,1,1,1,0,0,0]
				case 30:
					return [3,2,2,1,0,0,0]
			return [7,5,4,1,0,0,0] # Th
		case 91:
			return [7,5,4,2,0,0,0] # Pa
		case 92:
			match nc:
				case 14:
					return [2,1,1,2,0,0,0]
				case 32:
					return [3,2,2,2,0,0,0]
			return [7,5,4,2,0,0,0] # U


def NaturalAtomicOrbital(mo_mwfn, modify_minimal_shells = []):
	nao_mwfn = mo_mwfn.Clone()
	extra_info = dict()
	basis_indices_by_center = nao_mwfn.Atom2BasisList()
	shell_indices_by_center = nao_mwfn.Atom2ShellList()
	basis_indices_by_shell = nao_mwfn.Shell2BasisList()
	basis_indices_nmb, basis_indices_nrb = None, None
	if nao_mwfn.Overlap.shape != tuple([nao_mwfn.getNumBasis()] * 2):
		nao_mwfn.Overlap = eint.PyscfOverlap(nao_mwfn, nao_mwfn)
	S = nao_mwfn.Overlap
	angulars = [shell.Type for center in nao_mwfn.Centers for shell in center.Shells]
	minimal_shells = [MinimalShells(center.Index, round(center.Nuclear_charge)) for center in nao_mwfn.Centers]
	for icenter, minimal_shell in modify_minimal_shells:
		minimal_shells[icenter] = minimal_shell
	print("Natural atomic orbitals:")
	for spin in nao_mwfn.getSpins():
		D = nao_mwfn.getDensity(spin)
		basis_indices_nmb, basis_indices_nrb, W, N = generateNaturalAtomicOrbital(shell_indices_by_center, basis_indices_by_shell, basis_indices_by_center, angulars, D, S, minimal_shells)
		nao_mwfn.setOccupation(W, spin)
		nao_mwfn.setCoefficientMatrix(N, spin)
		nao_mwfn.setEnergy([0] * nao_mwfn.getNumIndBasis(), spin)
		match spin:
			case 0:
				extra_info["NAO_density_matrix"] = N.T @ S @ D @ S @ N
			case 1:
				extra_info["NAO_alpha_density_matrix"] = N.T @ S @ D @ S @ N
			case 2:
				extra_info["NAO_beta_density_matrix"] = N.T @ S @ D @ S @ N
	extra_info["NAO_minimal_basis_indices"] = basis_indices_nmb
	extra_info["NAO_Rydberg_basis_indices"] = basis_indices_nrb
	return nao_mwfn, extra_info



import numpy as np
import scipy.linalg as sl


def Lowdin(C, S, basis_indices_by_center):
	SsqrtC = sl.sqrtm(S) @ C
	Qs = []
	for basis_indices in basis_indices_by_center:
		tmp = SsqrtC[basis_indices, :]
		Qs.append(tmp.T @ tmp)
	return Qs

def Lowdin_func(C,S,basis_indices_by_center):
    nbasis=C.shape[0]
    norbitals=C.shape[1]
    assert (nbasis,nbasis)==S.shape,"Dimensions of coefficient and overlap matrices do not match!"
    S12C=sl.sqrtm(S)@C
    Qs=[]
    for basis_indices in basis_indices_by_center:
        S12Ca=S12C[basis_indices,:]
        Qa=S12Ca.T@S12Ca
        Qs.append(Qa)
    return Qs

def Lowdin_jac(C,C0,S,basis_indices_by_center):
    nbasis=C.shape[0]
    norbitals=C.shape[1]
    assert (nbasis,nbasis)==S.shape,"Dimensions of coefficient and overlap matrices do not match!"
    S12=sl.sqrtm(S)
    S12C0=S12@C0
    S12C=S12@C
    QrrUkrs=[]
    for basis_indices in basis_indices_by_center:
        S12C0a=S12C0[basis_indices,:]
        S12Ca=S12C[basis_indices,:]
        QrrUkrs.append(2*S12C0a.T@S12Ca)
    return QrrUkrs




import itertools

def FormatRange(numbers):
	if not numbers:
		return ""
	sorted_unique = sorted(set(numbers))
	groups = []
	for key, group in itertools.groupby(enumerate(sorted_unique), lambda x: x[1] - x[0]):
		group = list(group)
		start = group[0][1]
		end = group[-1][1]
		groups.append((start, end))
	parts = []
	for s, e in groups:
		parts.append(f"{s}-{e}" if s != e else f"{s}")
	return ",".join(parts)



import numpy as np
import copy as cp


class CubeTranslation:
    NumPoints=None
    Direction=None

class CubeAtom:
    Index=None
    Nuclear_charge=None
    Coordinates=None

class Cube:
    Title1=None
    Title2=None
    Origin=None
    Translations=None
    Atoms=None
    Data=None

    def __init__(self,filename):
        with open(filename,'r') as f:
            splitline=f.readline().split()
            splitline=f.readline().split()
            splitline=f.readline().split()
            natoms=int(splitline[0])
            self.Origin=np.array(splitline[1:],"float")
            self.Translations=[CubeTranslation() for i in range(3)]
            for translation in self.Translations:
                splitline=f.readline().split()
                translation.NumPoints=int(splitline[0])
                translation.Direction=np.array(splitline[1:],"float")
            self.Atoms=[CubeAtom() for i in range(natoms)]
            for atom in self.Atoms:
                splitline=f.readline().split()
                atom.Index=int(splitline[0])
                atom.Nuclear_charge=float(splitline[1])
                atom.Coordinates=np.array(splitline[2:],"float")
            npoints=self.Translations[0].NumPoints*self.Translations[1].NumPoints*self.Translations[2].NumPoints
            data=np.zeros(npoints)
            pointer=0
            while True:
                line=f.readline()
                if not line:
                    break
                splitline=line.split()
                nelements=len(splitline)
                data[pointer:pointer+nelements]=np.array(splitline,"float")
                pointer+=nelements
            self.Data=np.reshape(data,[self.Translations[0].NumPoints,self.Translations[1].NumPoints,self.Translations[2].NumPoints])

    def Differentiate(self):
        grad_dot_a=np.zeros_like(self.Data)
        grad_dot_a[0,:,:]=self.Data[1,:,:]-self.Data[0,:,:]
        grad_dot_a[-1,:,:]=self.Data[-1,:,:]-self.Data[-2,:,:]
        grad_dot_a[1:-1,:,:]=0.5*(self.Data[2:,:,:]-self.Data[0:-2,:,:])
        grad_dot_b=np.zeros_like(self.Data)
        grad_dot_b[:,0,:]=self.Data[:,1,:]-self.Data[:,0,:]
        grad_dot_b[:,-1,:]=self.Data[:,-1,:]-self.Data[:,-2,:]
        grad_dot_b[:,1:-1,:]=0.5*(self.Data[:,2:,:]-self.Data[:,0:-2,:])
        grad_dot_c=np.zeros_like(self.Data)
        grad_dot_c[:,:,0]=self.Data[:,:,1]-self.Data[:,:,0]
        grad_dot_c[:,:,-1]=self.Data[:,:,-1]-self.Data[:,:,-2]
        grad_dot_c[:,:,1:-1]=0.5*(self.Data[:,:,2:]-self.Data[:,:,0:-2])
        grad_norm=cp.deepcopy(self)
        grad_x=cp.deepcopy(self)
        grad_y=cp.deepcopy(self)
        grad_z=cp.deepcopy(self)
        A=np.zeros([3,3])
        for i in range(3):
            A[i,:]=self.Translations[i].Direction
        for i in range(self.Translations[0].NumPoints):
            for j in range(self.Translations[1].NumPoints):
                for k in range(self.Translations[2].NumPoints):
                    B=[grad_dot_a[i,j,k],grad_dot_b[i,j,k],grad_dot_c[i,j,k]]
                    grad_x.Data[i,j,k],grad_y.Data[i,j,k],grad_z.Data[i,j,k]=np.linalg.solve(A,B)
        grad_norm.Data=np.sqrt(grad_x.Data**2+grad_y.Data**2+grad_z.Data**2)
        return grad_norm,grad_x,grad_y,grad_z

    def Function(self,function):
        self.Data=np.zeros([self.Translations[i].NumPoints for i in range(3)])
        for i in range(self.Translations[0].NumPoints):
            for j in range(self.Translations[1].NumPoints):
                for k in range(self.Translations[2].NumPoints):
                    x,y,z=self.Origin+i*self.Translations[0].Direction+j*self.Translations[1].Direction+k*self.Translations[2].Direction
                    self.Data[i,j,k]=function(x,y,z)

    def Export(self,filename):
        with open(filename,'w') as f:
            f.write("Generated by Orbaplaw\n" if self.Title1 is None else (self.Title1+'\n'))
            f.write(("Total %d grids\n" % int(self.Data.size)) if self.Title2 is None else (self.Title2+'\n'))
            f.write("%d %f %f %f\n" % (len(self.Atoms),*self.Origin))
            for translation in self.Translations:
                f.write("%d %f %f %f\n" % (translation.NumPoints,*translation.Direction))
            for atom in self.Atoms:
                f.write("%d %f %f %f %f\n" % (atom.Index,atom.Nuclear_charge,*atom.Coordinates))
            for i in range(self.Translations[0].NumPoints):
                for j in range(self.Translations[1].NumPoints):
                    for k in range(self.Translations[2].NumPoints):
                        f.write("%13.5E" % self.Data[i,j,k])
                        if k%6==5 or k+1==self.Translations[2].NumPoints:
                            f.write('\n')




from setuptools import setup, find_packages

__version__ = "2.0.1"

setup(
		name = "Orbaplaw",
		version = __version__,
		author = "FreemanTheMaverick",
		description = "Orbital alignment analysis for plane wave basis sets",
		long_description = open("README.md").read(),
		long_description_content_type = "text/markdown",
		url = "https://github.com/FreemanTheMaverick/Orbaplaw.git",
		packages = find_packages(),
		entry_points = { "console_scripts": [
			"orbaplaw = Orbaplaw.main:main"
		]},
		install_requires = ["numpy", "scipy", "pyscf", "Maniverse", "libmwfn"],
		classifiers = ["Programming Language :: Python :: 3"]
)



import numpy as np


def ReadMwfnMat(file): # Very ugly codes to parse matrices in very ugly format. Do not try to understand.
    lines=None
    with open(file,'r') as f:
        lines=f.readlines()

    matrix_heads=[]
    for iline,line in enumerate(lines):
        if "**********" in line:
            matrix_heads.append(iline+1)
    matrix_texts=[]
    for imatrix,head in enumerate(matrix_heads):
        tail=len(lines) if head==matrix_heads[-1] else matrix_heads[imatrix+1]-2
        matrix_texts.append(lines[head:tail])
    matrices=[]
    for imatrix,matrix_text in enumerate(matrix_texts):
        nbasis=0
        block_heads=[]
        for iline,line in enumerate(matrix_text):
            if '.' not in line:
                block_heads.append(iline)
                nbasis=int(line.split()[-1])
        blocks=[]
        for jblock,head in enumerate(block_heads):
            tail=len(matrix_text) if head==block_heads[-1] else block_heads[jblock+1]
            blocks.append(matrix_text[head:tail])
        matrix=np.zeros([nbasis,nbasis])
        for jblock,block_text in enumerate(blocks):
            cols=None
            for kline,line in enumerate(block_text):
                if kline==0:
                    cols=np.array(block_text[0].split(),dtype="int")-1
                else:
                    row=int(line.split()[0])-1
                    nelements=len(line.split())-1
                    matrix[row,cols[:nelements]]=np.array(line.split()[1:])
        matrices.append(matrix+matrix.T-np.diag(np.diag(matrix)))
    return matrices



import numpy as np
import scipy.linalg as sl
from Orbaplaw import Integrals as eint


__angstrom2bohr__ = 1.8897259886

class MwfnCenter:
	Symbol = None # String
	Index = None # Int
	Nuclear_charge = None # Float
	Coordinates = None # np.array (1,3)
	Shells = None # MwfnShell
	def getNumShells(self):
		return len(self.Shells)

class MwfnShell:
	Type = None # int
	Center = None # MwfnCenter
	Exponents = None # List
	Coefficients = None # List
	def getSize(self):
		if self.Type >= 0:
			return ( self.Type + 1 ) * ( self.Type + 2 ) // 2
		else:
			return 2 * abs(self.Type) + 1
	def getNumPrims(self):
		return len(self.Exponents)
	
class MwfnOrbital:
	Type = None
	Energy = None
	Occ = None
	Sym = None
	Coeff = None # np.array (Nshell,1)

class MultiWaveFunction:

	Comment=""

	# Field 1
	Wfntype=None
	E_tot=None
	VT_ratio=None

	# Field 2
	Centers=None # List of MwfnCenter

	# Field 3
	Shells=None # List of MwfnShell

	# Field 4
	Orbitals=None # List of np.array

	# Field 5
	Total_density_matrix=None # np.array
	Alpha_density_matrix=None
	Beta_density_matrix=None
	Hamiltonian_matrix=None
	Alpha_Hamiltonian_matrix=None
	Beta_Hamiltonian_matrix=None
	Overlap_matrix=None
	Kinetic_energy_matrix=None
	Potential_energy_matrix=None
	X_electric_dipole_moment_matrix=None
	Y_electric_dipole_moment_matrix=None
	Z_electric_dipole_moment_matrix=None
	XX_electric_quadrupole_moment_matrix=None
	YY_electric_quadrupole_moment_matrix=None
	ZZ_electric_quadrupole_moment_matrix=None
	XY_electric_quadrupole_moment_matrix=None
	YZ_electric_quadrupole_moment_matrix=None
	XZ_electric_quadrupole_moment_matrix=None

	Extra_info={}

	def MatrixTransform(self):
		SPDFGHI=dict()
		SPDFGHI[0]=np.array([[1]])
		SPDFGHI[1]=np.eye(3)
		SPDFGHI[-2]=np.array([
			[0,0,1,0,0],
			[0,0,0,1,0],
			[0,1,0,0,0],
			[0,0,0,0,1],
			[1,0,0,0,0]])
		SPDFGHI[-3]=np.array([
			[0,0,0,1,0,0,0],
			[0,0,0,0,1,0,0],
			[0,0,1,0,0,0,0],
			[0,0,0,0,0,1,0],
			[0,1,0,0,0,0,0],
			[0,0,0,0,0,0,1],
			[1,0,0,0,0,0,0]])
		SPDFGHI[-4]=np.array([
			[0,0,0,0,1,0,0,0,0],
			[0,0,0,0,0,1,0,0,0],
			[0,0,0,1,0,0,0,0,0],
			[0,0,0,0,0,0,1,0,0],
			[0,0,1,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,1,0],
			[0,1,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,1],
			[1,0,0,0,0,0,0,0,0]])
		SPDFGHI[-5]=np.array([
			[0,0,0,0,0,1,0,0,0,0,0],
			[0,0,0,0,0,0,1,0,0,0,0],
			[0,0,0,0,1,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,1,0,0,0],
			[0,0,0,1,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,1,0,0],
			[0,0,1,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,1,0],
			[0,1,0,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,0,1],
			[1,0,0,0,0,0,0,0,0,0,0]])
		SPDFGHI[-6]=np.array([
			[0,0,0,0,0,0,1,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,1,0,0,0,0,0],
			[0,0,0,0,0,1,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,1,0,0,0,0],
			[0,0,0,0,1,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,1,0,0,0],
			[0,0,0,1,0,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,0,1,0,0],
			[0,0,1,0,0,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,0,0,1,0],
			[0,1,0,0,0,0,0,0,0,0,0,0,0],
			[0,0,0,0,0,0,0,0,0,0,0,0,1],
			[1,0,0,0,0,0,0,0,0,0,0,0,0]])
		nbasis=self.getNumBasis()
		mwfntransform=np.zeros([nbasis,nbasis])
		jbasis=0
		for center in self.Centers:
			for shell in center.Shells:
				l=shell.Type
				num=shell.getSize()
				mwfntransform[jbasis:jbasis+num,jbasis:jbasis+num]=SPDFGHI[l]
				jbasis+=num
		return mwfntransform

	def __init__(self,filename):
		def ReadMatrix(f,nrows,ncols,lower):
			matrix=np.zeros([nrows,ncols])
			assert nrows==ncols
			nelements=((1+nrows)*nrows//2) if lower else nrows*ncols
			elements=["114514" for i in range(nelements)]
			ielement=0
			finished=False
			while not finished:
				newline=f.readline()
				if not newline:
					break
				newvalues=newline.split()
				for newvalue in newvalues:
					try:
						elements[ielement]=float(newvalue)
					except ValueError:
						elements[ielement]=0.
					ielement+=1
					if ielement==nelements:
						finished=True
						break
			matrix=np.zeros([nrows,ncols])
			ielement=0
			for irow in range(nrows):
				for jcol in range(irow+1 if lower else ncols):
					matrix[irow,jcol]=elements[ielement]
					if lower:
						matrix[jcol,irow]=elements[ielement]
					ielement+=1
			return matrix

		with open(filename,'r') as f:
			nbasis=-114
			nindbasis=-114
			mwfntransform=None
			tmp_int=None
			while True:
				line=f.readline()
				if not line:
					break
				values=[]
				value=None
				if len(line.split())>1:
					values=line.split()
					value=values[1]

				if "# Comment: " in line:
					self.Comment=line[11:]
				
				# Field 1
				elif "Wfntype=" in line:
					self.Wfntype=int(value)
				elif "E_tot=" in line:
					self.E_tot=float(value)
				elif "VT_ratio=" in line:
					self.VT_ratio=float(value)

				# Field 2
				elif "Ncenter=" in line:
					self.Centers=[MwfnCenter() for i in range(int(value))]
				elif "$Centers" in line:
					for center in self.Centers:
						newline=f.readline()
						newvalues=newline.split()
						center.Symbol=newvalues[1]
						center.Index=int(newvalues[2])
						center.Nuclear_charge=float(newvalues[3])
						center.Coordinates=np.array(newvalues[4:7],dtype="float")*__angstrom2bohr__

				# Field 3
				elif "Nbasis=" in line:
					nbasis=int(value)
					if nindbasis!=114:
						self.Orbitals=[MwfnOrbital() for i in range(nindbasis*(1 if self.Wfntype==0 else 2))]
						for orbital in self.Orbitals:
							orbital.Coeff=np.zeros(nbasis)
				elif "Nindbasis=" in line:
					nindbasis=int(value)
					if nbasis!=114:
						self.Orbitals=[MwfnOrbital() for i in range(nindbasis*(1 if self.Wfntype==0 else 2))]
						for orbital in self.Orbitals:
							orbital.Coeff=np.zeros(nbasis)
				elif "Nshell=" in line:
					self.Shells=[MwfnShell() for i in range(int(value))]
				elif "$Shell types" in line:
					ishell=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							self.Shells[ishell].Type=int(newvalue)
							ishell+=1
							if ishell==len(self.Shells):
								finished=True
								break
				elif "$Shell centers" in line:
					ishell=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							self.Shells[ishell].Center=self.Centers[int(newvalue)-1]
							ishell+=1
							if ishell==len(self.Shells):
								finished=True
								break
					for center in self.Centers:
						center.Shells=[]
					for shell in self.Shells:
						shell.Center.Shells.append(shell)
				elif "$Shell contraction degrees" in line:
					ishell=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							self.Shells[ishell].Exponents=[0 for i in range(int(newvalue))]
							self.Shells[ishell].Coefficients=[0 for i in range(int(newvalue))]
							ishell+=1
							if ishell==len(self.Shells):
								finished=True
								break
				elif "$Primitive exponents" in line:
					ishell=0
					jprim=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							self.Shells[ishell].Exponents[jprim]=float(newvalue)
							jprim+=1
							if jprim==len(self.Shells[ishell].Exponents):
								ishell+=1
								jprim=0
							if ishell==len(self.Shells):
								finished=True
								break
				elif "$Contraction coefficients" in line:
					ishell=0
					jprim=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							self.Shells[ishell].Coefficients[jprim]=float(newvalue)
							jprim+=1
							if jprim==len(self.Shells[ishell].Coefficients):
								ishell+=1
								jprim=0
							if ishell==len(self.Shells):
								finished=True
								break

				# Field 4
				elif "Index=" in line:
					if mwfntransform is None:
						mwfntransform=self.MatrixTransform()
					tmp_int=int(value)-1
				elif "Type=" in line:
					self.Orbitals[tmp_int].Type=int(value)
				elif "Energy=" in line:
					self.Orbitals[tmp_int].Energy=float(value)
				elif "Occ=" in line:
					self.Orbitals[tmp_int].Occ=float(value)
				elif "Sym=" in line:
					self.Orbitals[tmp_int].Sym=value
				elif "$Coeff" in line:
					ibasis=0
					finished=False
					while not finished:
						newline=f.readline()
						if not newline:
							break
						newvalues=newline.split()
						for newvalue in newvalues:
							try:
								self.Orbitals[tmp_int].Coeff[ibasis]=float(newvalue)
							except ValueError:
								self.Orbitals[tmp_int].Coeff[ibasis]=0.
							ibasis+=1
							if ibasis==self.getNumBasis():
								finished=True
								break

				# Field 5
				elif "$Total density matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Total_density_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Alpha density matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Alpha_density_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Beta density matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Beta_density_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$1-e Hamiltonian matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Hamiltonian_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Alpha 1-e Hamiltonian matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Alpha_Hamiltonian_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Beta 1-e Hamiltonian matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Beta_Hamiltonian_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Overlap matrix" in line:
					nrows=int(values[3])
					ncols=int(values[4])
					lower=int(values[6])
					self.Overlap_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Kinetic energy matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Kinetic_energy_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Potential energy matrix" in line:
					nrows=int(values[4])
					ncols=int(values[5])
					lower=int(values[7])
					self.Potential_energy_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$X electric dipole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.X_electric_dipole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Y electric dipole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.Y_electric_dipole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$Z electric dipole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.Z_electric_dipole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$XX electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.XX_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$YY electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.YY_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$ZZ electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.ZZ_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$XY electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.XY_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$YZ electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.YZ_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
				elif "$XZ electric quadrupole moment matrix" in line:
					nrows=int(values[6])
					ncols=int(values[7])
					lower=int(values[9])
					self.XZ_electric_quadrupole_moment_matrix=mwfntransform.T@ReadMatrix(f,nrows,ncols,lower)@mwfntransform
		for spin in ([0] if self.Wfntype==0 else [1,2]):
			self.setCoefficientMatrix(spin,mwfntransform.T@self.getCoefficientMatrix(spin))
		Extra_info={}

	def getNumElec(self,spin):
		naelec=0.
		nbelec=0.
		for orbital in self.Orbitals:
			if orbital.Type==0:
				naelec+=orbital.Occ/2
				nbelec+=orbital.Occ/2
			elif orbital.Type==1:
				naelec+=orbital.Occ
			elif orbital.Type==2:
				nbelec+=orbital.Occ
		if spin==0:
			return naelec+nbelec
		elif spin==1:
			return naelec
		elif spin==2:
			return nbelec

	def getCharge(self):
		n=-self.getNumElec(0)
		for center in self.Centers:
			n+=center.Nuclear_charge
		return n

	def getSpin(self):
		return self.getNumElec(1)-self.getNumElec(2)

	def getNumCenters(self):
		return len(self.Centers)

	def getNumBasis(self):
		n=0
		for shell in self.Shells:
			n+=shell.getSize()
		return n

	def getNumIndBasis(self):
		return len(self.Orbitals)//(2 if self.Wfntype==1 else 1)

	def getNumPrims(self):
		n=0
		for shell in self.Shells:
			n+=(abs(shell.Type)+1)*(abs(shell.Type)+2)//2*len(shell.Exponents)
		return n

	def getNumShells(self):
		return len(self.Shells)

	def getNumPrimShells(self):
		n=0
		for shell in self.Shells:
			n+=len(shell.Exponents)
		return n

	def getShellIndexByCenter(self,centers=-1):
		indices=[]
		if centers==-1:
			for center in self.Centers:
				indices.append(self.getShellIndexByCenter(center))
			return indices
		elif type(centers) is list:
			for center in centers:
				indices.append(self.getShellIndexByCenter(center))
			return indices
		else:
			assert type(centers) is MwfnCenter or int
			thiscenter=None
			if type(centers) is MwfnCenter:
				assert centers in self.Centers
				thiscenter=centers
			else:
				thiscenter=self.Centers[centers]
			ishell=0
			for shell in self.Shells:
				if shell.Center is thiscenter:
					indices.append(ishell)
				ishell+=1
		return indices

	def getShellCenters(self):
		center_list=[]
		for shell in self.Shells:
			center_list.append(self.Centers.index(shell.Center))
		return center_list

	def getBasisIndexByCenter(self,centers=-1):
		indices=[]
		if centers==-1:
			for center in self.Centers:
				indices.append(self.getBasisIndexByCenter(center))
			return indices
		elif type(centers) is list:
			for center in centers:
				indices.append(self.getBasisIndexByCenter(center))
			return indices
		else:
			assert type(centers) is MwfnCenter or int
			thiscenter=None
			if type(centers) is MwfnCenter:
				assert centers in self.Centers
				thiscenter=centers
			else:
				thiscenter=self.Centers[centers]
			ibasis=0
			for shell in self.Shells:
				if shell.Center is thiscenter:
					for count in range(shell.getSize()):
						indices.append(ibasis)
						ibasis+=1
				else:
					ibasis+=shell.getSize()
		return indices

	def getBasisIndexByShell(self,shells=-1):
		indices=[]
		if shells==-1:
			for shell in self.Shells:
				indices.append(self.getBasisIndexByShell(shell))
			return indices
		elif type(shells) is list:
			for shell in shells:
				indices.append(self.getBasisIndexByShell(shell))
			return indices
		else:
			assert type(shells) is MwfnShell or int
			thisshell=None
			if type(shells) is MwfnShell:
				assert shells in self.Shells
				thisshell=shells
			else:
				thisshell=self.Shells[shells]
			ibasis=0
			for shell in self.Shells:
				for i in range(shell.getSize()):
					if thisshell==shell:
						indices.append(ibasis)
					ibasis+=1
			return indices

	def getCoefficientMatrix(self,type_):
		matrix=np.zeros([self.getNumBasis(),self.getNumIndBasis()])
		orbital_indices=[i for i in range(len(self.Orbitals)) if self.Orbitals[i].Type==type_]
		for i in range(len(orbital_indices)):
			matrix[:,i]=self.Orbitals[orbital_indices[i]].Coeff
		return matrix

	def setCoefficientMatrix(self,type_,matrix):
		orbital_indices=[i for i in range(len(self.Orbitals)) if self.Orbitals[i].Type==type_]
		assert matrix.shape==(self.getNumBasis(),len(orbital_indices))
		for i in range(len(orbital_indices)):
			self.Orbitals[orbital_indices[i]].Coeff=matrix[:,i]

	def getEnergy(self,type_):
		return [orbital.Energy for orbital in self.Orbitals if orbital.Type==type_]

	def setEnergy(self,type_,energies):
		orbital_indices=[i for i in range(len(self.Orbitals)) if self.Orbitals[i].Type==type_]
		assert len(energies)==len(orbital_indices)
		for i in range(len(orbital_indices)):
			self.Orbitals[orbital_indices[i]].Energy=energies[i]

	def calcOverlap(self):
		self.Overlap_matrix = eint.PyscfOverlap(self,self)

	def calcHamiltonian(self):
		S=self.Overlap_matrix
		if self.Wfntype==0:
			E=np.diag(self.getEnergy(0))
			C=self.getCoefficientMatrix(0)
			self.Hamiltonian_matrix=S@C@E@C.T@S
		elif self.Wfntype==1:
			Ea=np.diag(self.getEnergy(1))
			Ca=self.getCoefficientMatrix(1)
			self.Alpha_Hamiltonian_matrix=S@Ca@Ea@Ca.T@S
			Eb=np.diag(self.getEnergy(2))
			Cb=self.getCoefficientMatrix(2)
			self.Beta_Hamiltonian_matrix=S@Cb@Eb@Cb.T@S

	def getFock(self, spin):
		S = self.Overlap_matrix
		E = np.diag(self.getEnergy(spin))
		C = self.getCoefficientMatrix(spin)
		return S @ C @ E @ C.T @ S

	def getOccupation(self,type_):
		return [orbital.Occ for orbital in self.Orbitals if orbital.Type==type_]

	def setOccupation(self,type_,occupations):
		orbital_indices=[i for i in range(len(self.Orbitals)) if self.Orbitals[i].Type==type_]
		assert len(occupations)==len(orbital_indices)
		for i in range(len(orbital_indices)):
			self.Orbitals[orbital_indices[i]].Occ=occupations[i]

	def calcDensity(self):
		S=self.Overlap_matrix
		if self.Wfntype==0:
			N=np.diag(self.getOccupation(0))
			C=self.getCoefficientMatrix(0)
			self.Total_density_matrix=C@N@C.T
		elif self.Wfntype==1:
			Na=np.diag(self.getOccupation(1))
			Ca=self.getCoefficientMatrix(1)
			self.Alpha_density_matrix=Ca@Na@Ca.T
			Nb=np.diag(self.getOccupation(2))
			Cb=self.getCoefficientMatrix(2)
			self.Beta_density_matrix=Cb@Nb@Cb.T

	def GramSchmidt(self,spin,fix):
		S=self.Overlap_matrix
		C=self.getCoefficientMatrix(spin)
		Cfix=C[:,:fix]
		Ccng=C[:,fix:]
		np.random.seed(0)
		Ccng=np.random.rand(*Ccng.shape)
		Cnew=np.hstack([Cfix,Ccng])
		for i in range(fix,Cnew.shape[1]):
			vector=Cnew[:,i]*1
			for j in range(i):
				Cnew[:,i]-=(vector@S@Cnew[:,j])*Cnew[:,j]
			Cnew[:,i]/=np.sqrt(Cnew[:,i]@S@Cnew[:,i])
		self.setCoefficientMatrix(spin,Cnew)
				
	def Export(self,filename):

		def PrintMatrix(f,matrix,lower):
			for irow in range(matrix.shape[0]):
				for jcol in range(irow+1 if lower else matrix.shape[1]):
					f.write(" "+str(matrix[irow,jcol]))
				f.write("\n")

		with open(filename,'w') as f:
			f.write("# Generated by Orbaplaw\n")
			f.write("\n# Comment: "+self.Comment)

			# Field 1
			f.write("\n\n# Overview\n")
			f.write("Wfntype= "+str(self.Wfntype)+"\n")
			f.write("Charge= "+str(self.getCharge())+"\n")
			f.write("Naelec= "+str(self.getNumElec(1))+"\n")
			f.write("Nbelec= "+str(self.getNumElec(2))+"\n")
			f.write("E_tot= "+str(self.E_tot)+"\n")
			f.write("VT_ratio= "+str(self.VT_ratio)+"\n")

			# Field 2
			f.write("\n\n# Atoms\n")
			f.write("Ncenter= "+str(self.getNumCenters())+"\n")
			f.write("$Centers\n")
			for icenter,center in enumerate(self.Centers):
				f.write(str(icenter+1)+" ")
				f.write(str(center.Symbol)+" ")
				f.write(str(center.Index)+" ")
				f.write(str(center.Nuclear_charge)+" ")
				f.write(str(center.Coordinates[0]/__angstrom2bohr__)+" ")
				f.write(str(center.Coordinates[1]/__angstrom2bohr__)+" ")
				f.write(str(center.Coordinates[2]/__angstrom2bohr__)+"\n")

			# Field 3
			f.write("\n\n# Basis set\n")
			f.write("Nbasis= "+str(self.getNumBasis())+"\n")
			f.write("Nindbasis= "+str(self.getNumIndBasis())+"\n")
			f.write("Nprims= "+str(self.getNumPrims())+"\n")
			f.write("Nshell= "+str(self.getNumShells())+"\n")
			f.write("Nprimshell= "+str(self.getNumPrimShells())+"\n")
			f.write("$Shell types")
			thiscenter=None
			for shell in self.Shells:
				if thiscenter is not shell.Center:
					f.write("\n")
					thiscenter=shell.Center
				f.write(" "+str(shell.Type))
			f.write("\n$Shell centers")
			thiscenter=None
			for shell in self.Shells:
				if thiscenter is not shell.Center:
					f.write("\n")
					thiscenter=shell.Center
				f.write(" "+str(self.Centers.index(shell.Center)+1))
			f.write("\n$Shell contraction degrees")
			for shell in self.Shells:
				if thiscenter is not shell.Center:
					f.write("\n")
					thiscenter=shell.Center
				f.write(" "+str(shell.getNumPrims()))
			f.write("\n$Primitive exponents\n")
			for shell in self.Shells:
				for exponent in shell.Exponents:
					f.write(" "+str(exponent))
				f.write("\n")
			f.write("$Contraction coefficients\n")
			for shell in self.Shells:
				for coefficient in shell.Coefficients:
					f.write(" "+str(coefficient))
				f.write("\n")

			# Field 4
			mwfntransform=self.MatrixTransform()
			f.write("\n\n# Orbitals")
			for iorbital,orbital in zip(range(len(self.Orbitals)),self.Orbitals):
				f.write("\nIndex= %9d\n" % (iorbital+1))
				f.write("Type= "+str(orbital.Type)+"\n")
				f.write("Energy= "+str(orbital.Energy)+"\n")
				f.write("Occ= "+str(orbital.Occ)+"\n")
				f.write("Sym= "+str(orbital.Sym)+"\n")
				f.write("$Coeff\n")
				for element in mwfntransform@orbital.Coeff:
					f.write(" "+str(element))
				f.write("\n")

			# Field 5
			f.write("\n\n# Matrices\n")
			if self.Total_density_matrix is not None:
				f.write("$Total density matrix, dim= "+str(self.Total_density_matrix.shape[0])+" "+str(self.Total_density_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Total_density_matrix@mwfntransform.T,True)
			if self.Alpha_density_matrix is not None:
				f.write("$Alpha density matrix, dim= "+str(self.Alpha_density_matrix.shape[0])+" "+str(self.Alpha_density_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Alpha_density_matrix@mwfntransform.T,True)
			if self.Beta_density_matrix is not None:
				f.write("$Beta density matrix, dim= "+str(self.Beta_density_matrix.shape[0])+" "+str(self.Beta_density_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Beta_density_matrix@mwfntransform.T,True)
			if self.Hamiltonian_matrix is not None:
				f.write("$1-e Hamiltonian matrix, dim= "+str(self.Hamiltonian_matrix.shape[0])+" "+str(self.Hamiltonian_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Hamiltonian_matrix@mwfntransform.T,True)
			if self.Alpha_Hamiltonian_matrix is not None:
				f.write("$Alpha 1-e Hamiltonian matrix, dim= "+str(self.Alpha_Hamiltonian_matrix.shape[0])+" "+str(self.Alpha_Hamiltonian_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Alpha_Hamiltonian_matrix@mwfntransform.T,True)
			if self.Beta_Hamiltonian_matrix is not None:
				f.write("$Beta 1-e Hamiltonian matrix, dim= "+str(self.Beta_Hamiltonian_matrix.shape[0])+" "+str(self.Beta_Hamiltonian_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Beta_Hamiltonian_matrix@mwfntransform.T,True)
			if self.Overlap_matrix is not None:
				f.write("$Overlap matrix, dim= "+str(self.Overlap_matrix.shape[0])+" "+str(self.Overlap_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Overlap_matrix@mwfntransform.T,True)
			if self.Kinetic_energy_matrix is not None:
				f.write("$Kinetic energy matrix, dim= "+str(self.Kinetic_energy_matrix.shape[0])+" "+str(self.Kinetic_energy_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Kinetic_energy_matrix@mwfntransform.T,True)
			if self.Potential_energy_matrix is not None:
				f.write("$Potential energy matrix, dim= "+str(self.Potential_energy_matrix.shape[0])+" "+str(self.Potential_energy_matrix.shape[1])+" lower= 1\n")
				PrintMatrix(f,mwfntransform@self.Potential_energy_matrix@mwfntransform.T,True)
